@@ -115,16 +115,21 @@ static const struct vk_ext vk_device_extensions[] = {
     }
 };
 
+// Private struct for pl_vk_inst
+struct priv {
+    VkDebugReportCallbackEXT debug_cb;
+};
+
 void pl_vk_inst_destroy(const struct pl_vk_inst **inst_ptr)
 {
     const struct pl_vk_inst *inst = *inst_ptr;
     if (!inst)
         return;
 
-    VkDebugReportCallbackEXT debug_cb = (VkDebugReportCallbackEXT) inst->priv;
-    if (debug_cb) {
+    struct priv *p = TA_PRIV(inst);
+    if (p->debug_cb) {
         VK_LOAD_FUN(inst->instance, vkDestroyDebugReportCallbackEXT)
-        vkDestroyDebugReportCallbackEXT(inst->instance, debug_cb, VK_ALLOC);
+        vkDestroyDebugReportCallbackEXT(inst->instance, p->debug_cb, VK_ALLOC);
     }
 
     vkDestroyInstance(inst->instance, VK_ALLOC);
@@ -292,12 +297,15 @@ layers_done:
         vkCreateDebugReportCallbackEXT(inst, &dinfo, VK_ALLOC, &debug_cb);
     }
 
-    struct pl_vk_inst *pl_vk = talloc_struct(NULL, struct pl_vk_inst, {
+    struct pl_vk_inst *pl_vk = talloc_priv(NULL, struct pl_vk_inst, struct priv);
+    *pl_vk = (struct pl_vk_inst) {
         .instance = inst,
-        .priv = (uint64_t) debug_cb,
         .extensions = exts,
         .num_extensions = num_exts,
-    });
+    };
+
+    struct priv *p = TA_PRIV(pl_vk);
+    p->debug_cb = debug_cb;
 
     pl_vk->extensions = talloc_steal(pl_vk, pl_vk->extensions);
     talloc_free(tmp);
@@ -323,7 +331,7 @@ void pl_vulkan_destroy(const struct pl_vulkan **pl_vk)
 
     pl_gpu_destroy((*pl_vk)->gpu);
 
-    struct vk_ctx *vk = (*pl_vk)->priv;
+    struct vk_ctx *vk = TA_PRIV(*pl_vk);
     if (vk->dev) {
         PL_DEBUG(vk, "Flushing remaining commands...");
         vk_wait_idle(vk);
@@ -340,9 +348,11 @@ void pl_vulkan_destroy(const struct pl_vulkan **pl_vk)
     TA_FREEP((void **) pl_vk);
 }
 
-static bool supports_surf(struct vk_ctx *vk, VkPhysicalDevice physd,
+static bool supports_surf(struct pl_context *ctx, VkPhysicalDevice physd,
                           VkSurfaceKHR surf)
 {
+    struct { struct pl_context *ctx; } *vk = (void *) &ctx;
+
     uint32_t qfnum;
     vkGetPhysicalDeviceQueueFamilyProperties(physd, &qfnum, NULL);
 
@@ -357,17 +367,20 @@ error:
     return false;
 }
 
-static bool find_physical_device(struct vk_ctx *vk,
-                                 const struct pl_vulkan_params *params)
+VkPhysicalDevice pl_vulkan_choose_device(struct pl_context *ctx,
+                                         const struct pl_vulkan_device_params *params)
 {
+    struct { struct pl_context *ctx; } *vk = (void *) &ctx;
+
+    pl_assert(params->instance);
     PL_INFO(vk, "Probing for vulkan devices:");
-    bool ret = false;
+    VkPhysicalDevice dev = VK_NULL_HANDLE;
 
     VkPhysicalDevice *devices = NULL;
     uint32_t num = 0;
-    VK(vkEnumeratePhysicalDevices(vk->inst, &num, NULL));
+    VK(vkEnumeratePhysicalDevices(params->instance, &num, NULL));
     devices = talloc_array(NULL, VkPhysicalDevice, num);
-    VK(vkEnumeratePhysicalDevices(vk->inst, &num, devices));
+    VK(vkEnumeratePhysicalDevices(params->instance, &num, devices));
 
     static const struct { const char *name; int priority; } types[] = {
         [VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU]   = {"discrete",   5},
@@ -385,14 +398,14 @@ static bool find_physical_device(struct vk_ctx *vk,
         VkPhysicalDeviceType t = props.deviceType;
         PL_INFO(vk, "    GPU %d: %s (%s)", i, props.deviceName, types[t].name);
 
-        if (params->surface && !supports_surf(vk, devices[i], params->surface)) {
+        if (params->surface && !supports_surf(ctx, devices[i], params->surface)) {
             PL_DEBUG(vk, "      -> excluding due to lack of surface support");
             continue;
         }
 
         if (params->device_name && params->device_name[0] != '\0') {
             if (strcmp(params->device_name, props.deviceName) == 0) {
-                vk->physd = devices[i];
+                dev = devices[i];
                 best = 10; // high number...
             } else {
                 PL_DEBUG(vk, "      -> excluding due to name mismatch");
@@ -406,21 +419,14 @@ static bool find_physical_device(struct vk_ctx *vk,
         }
 
         if (types[t].priority > best) {
-            vk->physd = devices[i];
+            dev = devices[i];
             best = types[t].priority;
         }
     }
 
-    if (!vk->physd) {
-        PL_FATAL(vk, "Found no suitable device, giving up.");
-        goto error;
-    }
-
-    ret = true;
-
 error:
     talloc_free(devices);
-    return ret;
+    return dev;
 }
 
 // Find the most specialized queue supported a combination of flags. In cases
@@ -519,16 +525,21 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         }
     }
 
-    if (idx_tf >= 0 && idx_tf != idx_gfx)
-        PL_INFO(vk, "Using async transfer (QF %d)", idx_tf);
-    if (idx_comp >= 0 && idx_comp != idx_gfx)
-        PL_INFO(vk, "Using async compute (QF %d)", idx_comp);
-
     // Fall back to supporting compute shaders via the graphics pool for
     // devices which support compute shaders but not async compute.
     if (idx_comp < 0 && qfs[idx_gfx].queueFlags & VK_QUEUE_COMPUTE_BIT)
         idx_comp = idx_gfx;
 
+    if (params->blacklist_caps & PL_GPU_CAP_COMPUTE) {
+        PL_INFO(vk, "Disabling compute shaders (blacklisted)");
+        idx_comp = -1;
+    }
+
+    if (idx_tf >= 0 && idx_tf != idx_gfx)
+        PL_INFO(vk, "Using async transfer (QF %d)", idx_tf);
+    if (idx_comp >= 0 && idx_comp != idx_gfx)
+        PL_INFO(vk, "Using async compute (QF %d)", idx_comp);
+
     // Cache the transfer queue alignment requirements
     if (idx_tf >= 0)
         vk->transfer_alignment = qfs[idx_tf].minImageTransferGranularity;
@@ -554,7 +565,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     const char ***exts = &vk->exts;
     int *num_exts = &vk->num_exts;
     if (params->surface)
-        TARRAY_APPEND(vk, *exts, *num_exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
+        TARRAY_APPEND(vk->ta, *exts, *num_exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
 
     // Keep track of function pointers we need to load
     const struct vk_ext_fun **ext_funs = NULL;
@@ -567,7 +578,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         const struct vk_ext *ext = &vk_device_extensions[i];
         for (int n = 0; n < num_exts_avail; n++) {
             if (strcmp(ext->name, exts_avail[n].extensionName) == 0) {
-                TARRAY_APPEND(vk, *exts, *num_exts, ext->name);
+                TARRAY_APPEND(vk->ta, *exts, *num_exts, ext->name);
                 for (const struct vk_ext_fun *f = ext->funs; f->name; f++)
                     TARRAY_APPEND(tmp, ext_funs, num_ext_funs, f);
                 break;
@@ -577,14 +588,14 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
 
     // Add extra user extensions
     for (int i = 0; i < params->num_extensions; i++)
-        TARRAY_APPEND(vk, *exts, *num_exts, params->extensions[i]);
+        TARRAY_APPEND(vk->ta, *exts, *num_exts, params->extensions[i]);
 
     // Add optional extra user extensions
     for (int i = 0; i < params->num_opt_extensions; i++) {
         const char *ext = params->opt_extensions[i];
         for (int n = 0; n < num_exts_avail; n++) {
             if (strcmp(ext, exts_avail[n].extensionName) == 0) {
-                TARRAY_APPEND(vk, *exts, *num_exts, ext);
+                TARRAY_APPEND(vk->ta, *exts, *num_exts, ext);
                 break;
             }
         }
@@ -631,7 +642,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         struct vk_cmdpool *pool = vk_cmdpool_create(vk, qinfos[i], qfs[qf]);
         if (!pool)
             goto error;
-        TARRAY_APPEND(vk, vk->pools, vk->num_pools, pool);
+        TARRAY_APPEND(vk->ta, vk->pools, vk->num_pools, pool);
 
         // Update the pool_* pointers based on the corresponding index
         if (qf == idx_gfx)
@@ -648,6 +659,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
 error:
     PL_FATAL(vk, "Failed creating logical device!");
     talloc_free(tmp);
+    vk->failed = true;
     return false;
 }
 
@@ -655,8 +667,9 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
                                          const struct pl_vulkan_params *params)
 {
     params = PL_DEF(params, &pl_vulkan_default_params);
-    struct pl_vulkan *pl_vk = talloc_zero(NULL, struct pl_vulkan);
-    struct vk_ctx *vk = pl_vk->priv = talloc_zero(pl_vk, struct vk_ctx);
+    struct pl_vulkan *pl_vk = talloc_zero_priv(NULL, struct pl_vulkan, struct vk_ctx);
+    struct vk_ctx *vk = TA_PRIV(pl_vk);
+    vk->ta = pl_vk;
     vk->ctx = ctx;
     vk->inst = params->instance;
 
@@ -674,8 +687,18 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
     if (params->device) {
         PL_DEBUG(vk, "Using specified VkPhysicalDevice");
         vk->physd = params->device;
-    } else if (!find_physical_device(vk, params)) {
-        goto error;
+    } else {
+        vk->physd = pl_vulkan_choose_device(ctx, &(struct pl_vulkan_device_params) {
+            .instance       = vk->inst,
+            .surface        = params->surface,
+            .device_name    = params->device_name,
+            .allow_software = params->allow_software,
+        });
+
+        if (!vk->physd) {
+            PL_FATAL(vk, "Found no suitable device, giving up.");
+            goto error;
+        }
     }
 
     VkPhysicalDeviceProperties prop;
@@ -700,6 +723,11 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
     if (!pl_vk->gpu)
         goto error;
 
+    // Blacklist / restrict features
+    pl_gpu_caps *caps = (pl_gpu_caps*) &pl_vk->gpu->caps;
+    *caps &= ~(params->blacklist_caps);
+    vk->disable_events = params->disable_events;
+
     // Expose the resulting vulkan objects
     pl_vk->instance = vk->inst;
     pl_vk->phys_device = vk->physd;
@@ -721,5 +749,6 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
 error:
     PL_FATAL(vk, "Failed initializing vulkan device");
     pl_vulkan_destroy((const struct pl_vulkan **) &pl_vk);
+    vk->failed = true;
     return NULL;
 }
