@@ -27,12 +27,14 @@ struct priv {
     VkSurfaceKHR surf;
 
     // current swapchain and metadata:
+    struct pl_vulkan_swapchain_params params;
     VkSwapchainCreateInfoKHR protoInfo; // partially filled-in prototype
     VkSwapchainKHR swapchain;
     VkSwapchainKHR old_swapchain;
     int cur_width, cur_height;
     int swapchain_depth;
     int frames_in_flight;   // number of frames currently queued
+    bool suboptimal;        // true once VK_SUBOPTIMAL_KHR is returned
     struct pl_color_repr color_repr;
     struct pl_color_space color_space;
 
@@ -165,6 +167,12 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
     formats = talloc_array(NULL, VkSurfaceFormatKHR, num);
     VK(vkGetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, formats));
 
+    PL_DEBUG(gpu, "Available surface formats:");
+    for (int i = 0; i < num; i++) {
+        PL_DEBUG(gpu, "    %d: format: 0x%x, space: 0x%x", i,
+                 formats[i].format, formats[i].colorSpace);
+    }
+
     for (int i = 0; i < num; i++) {
         // A value of VK_FORMAT_UNDEFINED means we can pick anything we want
         if (formats[i].format == VK_FORMAT_UNDEFINED) {
@@ -214,15 +222,15 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
 
         // Make sure we can wrap this format to a meaningful, valid pl_fmt
         for (int n = 0; n < gpu->num_formats; n++) {
-            const struct pl_fmt *rafmt = gpu->formats[n];
-            const struct vk_format *vkfmt = rafmt->priv;
-            if (vkfmt->ifmt != formats[i].format)
+            const struct pl_fmt *plfmt = gpu->formats[n];
+            const struct vk_format **pvkfmt = TA_PRIV(plfmt);
+            if ((*pvkfmt)->tfmt != formats[i].format)
                 continue;
 
             enum pl_fmt_caps render_caps = 0;
             render_caps |= PL_FMT_CAP_RENDERABLE;
             render_caps |= PL_FMT_CAP_BLITTABLE;
-            if ((rafmt->caps & render_caps) != render_caps)
+            if ((plfmt->caps & render_caps) != render_caps)
                 continue;
 
             // format valid, use it
@@ -242,7 +250,7 @@ error:
 const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *plvk,
                               const struct pl_vulkan_swapchain_params *params)
 {
-    struct vk_ctx *vk = plvk->priv;
+    struct vk_ctx *vk = TA_PRIV(plvk);
     const struct pl_gpu *gpu = plvk->gpu;
 
     VkSurfaceFormatKHR sfmt = params->surface_format;
@@ -250,12 +258,13 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
     if (!pick_surf_format(gpu, vk, params->surface, &sfmt, &csp))
         return NULL;
 
-    struct pl_swapchain *sw = talloc_zero(NULL, struct pl_swapchain);
+    struct pl_swapchain *sw = talloc_zero_priv(NULL, struct pl_swapchain, struct priv);
     sw->impl = &vulkan_swapchain;
     sw->ctx = vk->ctx;
     sw->gpu = gpu;
 
-    struct priv *p = sw->priv = talloc_zero(sw, struct priv);
+    struct priv *p = TA_PRIV(sw);
+    p->params = *params;
     p->vk = vk;
     p->surf = params->surface;
     p->swapchain_depth = PL_DEF(params->swapchain_depth, 3);
@@ -310,7 +319,7 @@ error:
 static void vk_sw_destroy(const struct pl_swapchain *sw)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
     pl_gpu_flush(gpu);
@@ -328,7 +337,7 @@ static void vk_sw_destroy(const struct pl_swapchain *sw)
 
 static int vk_sw_latency(const struct pl_swapchain *sw)
 {
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     return p->swapchain_depth;
 }
 
@@ -395,6 +404,11 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     if (caps.maxImageCount)
         info->minImageCount = PL_MIN(info->minImageCount, caps.maxImageCount);
 
+    PL_DEBUG(vk, "Requested image size: %dx%d (min %dx%d < cur %dx%d < max %dx%d)",
+             w, h, caps.minImageExtent.width, caps.minImageExtent.height,
+             caps.currentExtent.width, caps.currentExtent.height,
+             caps.maxImageExtent.width, caps.maxImageExtent.height);
+
     // Default the requested size based on the reported extent
     if (caps.currentExtent.width != 0xFFFFFFFF)
         w = PL_DEF(w, caps.currentExtent.width);
@@ -456,7 +470,7 @@ static void destroy_swapchain(struct vk_ctx *vk, struct priv *p)
 static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
     VkImage *vkimages = NULL;
@@ -480,6 +494,7 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 
     VK(vkCreateSwapchainKHR(vk->dev, &sinfo, VK_ALLOC, &p->swapchain));
 
+    p->suboptimal = false;
     p->cur_width = sinfo.imageExtent.width;
     p->cur_height = sinfo.imageExtent.height;
 
@@ -505,8 +520,8 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
         VK(vkCreateSemaphore(vk->dev, &seminfo, VK_ALLOC, &sem_out));
 
         int idx = p->num_sems++;
-        TARRAY_GROW(p, p->sems_in, idx);
-        TARRAY_GROW(p, p->sems_out, idx);
+        TARRAY_GROW(sw, p->sems_in, idx);
+        TARRAY_GROW(sw, p->sems_out, idx);
         p->sems_in[idx] = sem_in;
         p->sems_out[idx] = sem_out;
     }
@@ -516,7 +531,7 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
         pl_tex_destroy(gpu, &p->images[i]);
 
     p->num_images = num_images;
-    TARRAY_GROW(p, p->images, num_images);
+    TARRAY_GROW(sw, p->images, num_images);
     for (int i = 0; i < num_images; i++) {
         const VkExtent2D *ext = &sinfo.imageExtent;
         p->images[i] = pl_vulkan_wrap(gpu, vkimages[i], ext->width, ext->height,
@@ -545,20 +560,29 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 error:
     PL_ERR(vk, "Failed (re)creating swapchain!");
     talloc_free(vkimages);
-    vkDestroySwapchainKHR(vk->dev, p->swapchain, VK_ALLOC);
-    p->swapchain = VK_NULL_HANDLE;
-    p->cur_width = p->cur_height = 0;
+    if (p->swapchain != sinfo.oldSwapchain) {
+        vkDestroySwapchainKHR(vk->dev, p->swapchain, VK_ALLOC);
+        p->swapchain = VK_NULL_HANDLE;
+        p->cur_width = p->cur_height = 0;
+        p->suboptimal = false;
+    }
     return false;
 }
 
 static bool vk_sw_start_frame(const struct pl_swapchain *sw,
                               struct pl_swapchain_frame *out_frame)
 {
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     struct vk_ctx *vk = p->vk;
     if (!p->swapchain && !vk_sw_recreate(sw, 0, 0))
         return false;
 
+    if (p->suboptimal && !p->params.allow_suboptimal) {
+        PL_TRACE(vk, "Swapchain is suboptimal, recreating...");
+        if (!vk_sw_recreate(sw, 0, 0))
+            return false;
+    }
+
     VkSemaphore sem_in = p->sems_in[p->idx_sems];
     PL_TRACE(vk, "vkAcquireNextImageKHR signals %p", (void *) sem_in);
 
@@ -568,6 +592,9 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
                                              sem_in, VK_NULL_HANDLE, &imgidx);
 
         switch (res) {
+        case VK_SUBOPTIMAL_KHR:
+            p->suboptimal = true;
+            // fall through
         case VK_SUCCESS:
             p->last_imgidx = imgidx;
             pl_vulkan_release(sw->gpu, p->images[imgidx],
@@ -580,7 +607,6 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
             };
             return true;
 
-        case VK_SUBOPTIMAL_KHR:
         case VK_ERROR_OUT_OF_DATE_KHR: {
             // In these cases try recreating the swapchain
             if (!vk_sw_recreate(sw, 0, 0))
@@ -607,7 +633,7 @@ static void present_cb(struct priv *p, void *arg)
 static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     struct vk_ctx *vk = p->vk;
     if (!p->swapchain)
         return false;
@@ -630,6 +656,8 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
     if (!vk_flush_commands(vk))
         return false;
 
+    vk_rotate_queues(p->vk);
+
     // Older nvidia drivers can spontaneously combust when submitting to the
     // same queue as we're rendering from, in a multi-queue scenario. Safest
     // option is to flush the commands first and then submit to the next queue.
@@ -649,8 +677,10 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
     PL_TRACE(vk, "vkQueuePresentKHR waits on %p", (void *) sem_out);
     VkResult res = vkQueuePresentKHR(queue, &pinfo);
     switch (res) {
-    case VK_SUCCESS:
     case VK_SUBOPTIMAL_KHR:
+        p->suboptimal = true;
+        // fall through
+    case VK_SUCCESS:
         return true;
 
     case VK_ERROR_OUT_OF_DATE_KHR:
@@ -667,7 +697,7 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
 
 static void vk_sw_swap_buffers(const struct pl_swapchain *sw)
 {
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
 
     while (p->frames_in_flight >= p->swapchain_depth)
         vk_poll_commands(p->vk, UINT64_MAX);
@@ -675,7 +705,7 @@ static void vk_sw_swap_buffers(const struct pl_swapchain *sw)
 
 static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
 {
-    struct priv *p = sw->priv;
+    struct priv *p = TA_PRIV(sw);
     bool ok = true;
 
     if ((*width && *width != p->cur_width) || (*height && *height != p->cur_height))
@@ -686,6 +716,12 @@ static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
     return ok;
 }
 
+bool pl_vulkan_swapchain_suboptimal(const struct pl_vulkan *vk)
+{
+    struct priv *p = TA_PRIV(vk);
+    return p->suboptimal;
+}
+
 static struct pl_sw_fns vulkan_swapchain = {
     .destroy      = vk_sw_destroy,
     .latency      = vk_sw_latency,
